#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>

#include "main.h"
#include <util/delay.h>
#include "globals.h"
#include "functions.h"
#include "cmd_terminal.h"
#include "uart.h"
#include "serial_flash.h"
#include "tlc5940.h"


// UART Terminal
// evaluate commands

void cmd_terminal() {
    uint8_t len;
    uint8_t mode_error=0;

    len=strlen((char*) gterminal_buffer);

    // commands in all modes
    if (gterminal_buffer[0] == 'C') {
        switch (gterminal_buffer[2]) {
    		case 'I':	cmd_get_ID((char*) &gterminal_buffer[2], len-2); break;
    		case 'V':	cmd_get_version((char*) &gterminal_buffer[2], len-2); break;
    		case 'L':	cmd_dump_events((char*) &gterminal_buffer[2], len-2); break;
    		case 'M':	cmd_set_mode((char*) &gterminal_buffer[2], len-2); break;
    		case 'Q':	cmd_reset_module((char*) &gterminal_buffer[2], len-2); break;
            default :	mode_error=1; break;
    	}
    } else {
        mode_error = 1;
    }

    if (mode_error) {
        mode_error=0;
        if (gconfig.mode == MODE_CONFIG) {
            // commands in config mode only
            if (gterminal_buffer[0] == 'C') {
                switch (gterminal_buffer[2]) {
            		case 'S':	cmd_event_store((char*) &gterminal_buffer[2], len-2); break;
            		case 'C':	cmd_event_clear((char*) &gterminal_buffer[2], len-2); break;	
                    case 'Z':	cmd_event_cfg_clear((char*) &gterminal_buffer[2], len-2); break;
                    default :	mode_error=1; break;
            	}
            } else if (gterminal_buffer[0] == 'E') {
                cmd_event_def((char*) gterminal_buffer, len);
            } else if (gterminal_buffer[0] == 'S') {
                switch (gterminal_buffer[2]) {
            		case 'W':	cmd_set_PWM((char*) &gterminal_buffer[2], len-2); break;
                    case 'I':	cmd_set_Imax((char*) &gterminal_buffer[2], len-2); break;
                    case 'P':	cmd_set_debounce((char*) &gterminal_buffer[2], len-2); break;
                    default :	mode_error=1; break;
            	}
            } else {
                mode_error = 1;
            }
        } else {
            mode_error = 1;
        }
    }

    if (mode_error) {
        cmd_error(PSTR("Unknown command"));
    }
    gterminal_buffer[0]=0;

}

// error, print ? + error messeage

void cmd_error(PGM_P error) {
	PRINT("? %S error ?\r\n", error);
}

// command successful, print "OK"

void cmd_OK() {
	uart_puts_P(PSTR("OK\r\n"));
}

// get ID

void cmd_get_ID(char * cmd_buffer, uint8_t len) {

	if (len!=1){				// check cmd length
		cmd_error(PSTR("Parameter"));
		return;
	}

	uart_puts_P(PSTR("LED\r\n"));
}

// get version

void cmd_get_version(char * cmd_buffer, uint8_t len) {

	if (len!=1){				// check cmd length
		cmd_error(PSTR("Parameter"));
		return;
	}

    PRINT("%01d.%02d\r\n", VERSION_MAJOR, VERSION_MINOR);
}

// define event

void cmd_event_def(char* terminal_buffer, uint8_t len) {

    uint8_t c;
    uint16_t v;
    char* next;

    next = strtok_P(terminal_buffer, PSTR(","));    //

    next = strtok_P(NULL, PSTR(","));
    if (next == NULL) {
        cmd_error(PSTR("Parameter"));
        return;
    }
    c = atoi(next);       // channel
    c = led_map(c);       // convert to internal number scheme

    next = strtok_P(NULL, PSTR(","));
    if (next == NULL) {
        cmd_error(PSTR("Parameter"));
        return;
    }
    v=atoi(next);       // PWM value

    // debug

    //PRINT("DEBUG c: %d v: %d\r\n" ,c , v);

    if (c>15 || v>4095) {
        cmd_error(PSTR("Parameter"));
        return;
    }
    
    gevent_data[c] = v;
    cmd_OK();
};

// store event in FLASH

void cmd_event_store(char* terminal_buffer, uint8_t len) {

    uint16_t a, i;
    char* next;

    next = strtok_P(terminal_buffer, PSTR(","));

    next = strtok_P(NULL, PSTR(","));
    if (next == NULL) {
        //PRINT("parameter error");       // Debug
        cmd_error(PSTR("Parameter"));
        return;
    }
    a=atoi(next);       // event

    if (a > 255) {
        //PRINT("value error"); //Debug
        cmd_error(PSTR("Parameter"));
        return;
    }

    // DEBUG
  /*
    for(i=0; i<16; i++) {
        PRINT("%2d, %4X\r\n", i, gevent_data[i]);
    }
  */
    // wait for flash to be ready, timeout after 50ms

    for(i=50; flash_busy() && i>0; i--) {
        _delay_ms(1);
        //PRINT(".");   // DEBUG
    }

    if (i) {
        flash_write_page((uint8_t*)gevent_data, a, 32);
        cmd_OK();
    } else {
        cmd_error(PSTR("Flash write"));
    }

    clear_event();
};

// clear event, all channels NOP

void cmd_event_clear(char* terminal_buffer, uint8_t len) {

    if (len!=1) {				// check cmd length
		cmd_error(PSTR("Parameter"));
		return;
	}

    clear_event();
    cmd_OK();
};

// dump event list

void cmd_dump_events(char* terminal_buffer, uint8_t len) {

    uint16_t i, j;
    uint16_t tmp[16];

    if (len!=1){				// check cmd length
		cmd_error(PSTR("Parameter"));
		return;
	}

    PRINT("Event list\r\n  No    C0   C1   C2   C3   C4   C5   C6   C7   C8   C9  C10  C11  C12  C13  C14  C15\r\n");
    for (i=0; i<256; i++) {
        PRINT(" %3d ", i);
        flash_read_page((uint8_t*)tmp, i, 32);
        for (j=0; j<16; j++) {
            if (tmp[j] == NOP_LED) {
                PRINT("     ");
            } else {
                PRINT(" %4u", tmp[j] & 0xFFF);
            }
        }
        PRINT("\r\n");
    }
};

// set PWM frequency and resolution

void cmd_set_PWM(char* terminal_buffer, uint8_t len) {

    uint16_t pwm;
    char* next;

    next = strtok_P(terminal_buffer, PSTR(","));

    next = strtok_P(NULL, PSTR(","));
    if (next == NULL) {
        cmd_error(PSTR("Parameter"));
        return;
    }
    pwm=atoi(next);       // pwm frequency

    // debug

    //PRINT("DEBUG c: %d v: %ld\r\n" ,c , v);

    switch (pwm) {
        case 1000: gconfig.pwm_prescaler_mask = 7; break;
        case 2000: gconfig.pwm_prescaler_mask = 3; break;
        case 4000: gconfig.pwm_prescaler_mask = 1; break;
        case 8000: gconfig.pwm_prescaler_mask = 0; break;
        default  : cmd_error(PSTR("Parameter")); return;
    }

    cfg_write();
    set_TLC_clock(gconfig.pwm_prescaler_mask);
    cmd_OK();  
};

// set max. LED current

void cmd_set_Imax(char* terminal_buffer, uint8_t len) {

    uint8_t v;
    char* next;

    next = strtok_P(terminal_buffer, PSTR(","));

    next = strtok_P(NULL, PSTR(","));
    if (next == NULL) {
        cmd_error(PSTR("Parameter"));
        return;
    }
    v=atoi(next);       // code for LED current

    // debug

    //PRINT("DEBUG %d v: %d\r\n", v);

    if(v>15) {
        cmd_error(PSTR("Parameter"));
        return;
    }

    gconfig.imax= v<<4;
    cfg_write();
    TLC5940_update_GS(gTLC5940_data_GS);
    gflag_new_event=1;
    cmd_OK();
};

// set mode

void cmd_set_mode(char* terminal_buffer, uint8_t len) {
    char a=terminal_buffer[2];
    uint8_t new_mode;

    switch(a) {
        case 'B':
            new_mode = MODE_BUS;
            MUX_BUS
        break;

        case 'E':
            new_mode = MODE_EVENT;
            MUX_STANDALONE
        break;
        
        case 'K':
            new_mode = MODE_CONFIG;
            MUX_STANDALONE
        break;

        default:
            cmd_error(PSTR("Parameter"));
            return; 
    }

    gconfig.mode = new_mode;
    cfg_write();
    cmd_OK();
};

// set debounce time

void cmd_set_debounce(char* terminal_buffer, uint8_t len) {

    uint32_t a;
    char* next;

    next = strtok_P(terminal_buffer, PSTR(","));

    next = strtok_P(NULL, PSTR(","));
    if (next == NULL) {
        // PRINT("VALUE");     // Debug
        cmd_error(PSTR("Parameter"));
        return;
    }
    a=atol(next);       // debounce

    if (a > 65535) {
        //PRINT("RANGE");     // Debug
        cmd_error(PSTR("Parameter"));
        return;
    }
    
    gconfig.debounce = a;
    cfg_write();
    cmd_OK();
};


// reset event config

void cmd_event_cfg_clear(char* terminal_buffer, uint8_t len) {

    uint16_t i, j;

    clear_event();
    for (i=0; i<256; i++) {
       
        // wait for flash to be ready, timeout after 50ms
        for(j=50; flash_busy() && j>0; j--) {
            _delay_ms(1);
            //PRINT(".");   // DEBUG
        }

        if (j) {
            flash_write_page((uint8_t*)gevent_data, i, 32);
        } else {
            cmd_error(PSTR("Flash erase"));
            return;
        }
    }
    cmd_OK();
}

void cmd_reset_module(char* terminal_buffer, uint8_t len) {
    uint8_t i;

    for(i=0; i<24; i++) {
        gTLC5940_data_GS[i]=0;
    }

    gconfig.pwm_prescaler_mask = 7;        // 1kHz
    gconfig.debounce = 0;
    gconfig.imax = 0;

    cfg_write();

    set_TLC_clock(gconfig.pwm_prescaler_mask);
    TLC5940_update_GS(gTLC5940_data_GS);
    gflag_new_event=1;
    cmd_OK();
}
